"""Contain the Keypair class and related methods"""

from hashlib import sha256

from kin_base.keypair import Keypair as BaseKeypair
from kin_base.exceptions import StellarSecretInvalidError
from kin_base.stellarxdr.StellarXDR_type import DecoratedSignature

from .utils import is_valid_secret_key

from typing import Optional


class Keypair:
    """A simpler version of kin_base.Keypair that caches the public address and secret seed."""

    def __init__(self, seed: Optional[str] = None):
        """
        # Create an instance of Keypair.

        :param seed: (Optional) The secret seed of an account
        """
        self.secret_seed = seed or self.generate_seed()
        if not is_valid_secret_key(self.secret_seed):
            raise StellarSecretInvalidError('invalid seed {}'.format(self.secret_seed))
        base_keypair = BaseKeypair.from_seed(self.secret_seed)
        self.public_address = base_keypair.address().decode()

        # Hint and signing key are needed to sign the tx
        self._hint = base_keypair.signature_hint()
        self._signing_key = base_keypair.signing_key

    def sign(self, data: bytes) -> DecoratedSignature:
        """
        Sign any data using the keypair's private key

        :param data: any data to sign
        :return: a decorated signature
        """
        signature = self._signing_key.sign(data)
        return DecoratedSignature(self._hint, signature)

    @staticmethod
    def address_from_seed(seed: str) -> str:
        """
        Get a public address from a secret seed.

        :param seed: The secret seed of an account.
        :return: A public address.
        """
        return BaseKeypair.from_seed(seed).address().decode()

    @staticmethod
    def generate_seed() -> str:
        """
        Generate a random secret seed.

        :return: A secret seed.
        """
        return BaseKeypair.random().seed().decode()

    @staticmethod
    def generate_hd_seed(base_seed: str, salt: str) -> str:
        """
        Generate a highly deterministic seed from a base seed + salt

        :param base_seed: The base seed to generate a seed from
        :param salt: A unique string that will be used to generate the seed
        :return: a new seed.
        """
        # Create a new raw seed from this hash
        raw_seed = sha256((base_seed + salt).encode()).digest()
        return BaseKeypair.from_raw_seed(raw_seed).seed().decode()

    def __str__(self):
        string_representation = '<Kin Keypair>: Address: {}, Seed: {}'.format(self.public_address, self.secret_seed)
        return string_representation



"""Contains errors related to the Kin SDK"""

from aiohttp.client_exceptions import ClientError

from kin_base.exceptions import HorizonError
from kin_base.exceptions import NotValidParamError, StellarAddressInvalidError, StellarSecretInvalidError

from .blockchain.errors import *


# All exceptions should subclass from SdkError in this module.
class SdkError(Exception):
    """Base class for all SDK errors."""

    def __init__(self, message=None, error_code=None, extra=None):
        super(SdkError, self).__init__(self)
        self.message = message or 'unknown error'
        self.error_code = error_code
        self.extra = dict(extra or ())

    def __str__(self):
        sb = list()
        sb.append("\n\tmessage='{}'".format(self.message))
        sb.append("\n\terror_code='{}'".format(self.error_code))
        sb.append("\n\textra data:")
        for key in self.extra:
            sb.append("\n\t\t{}='{}'".format(key, self.extra[key]))
        return ''.join(sb)

    def __repr__(self):
        return self.__class__.__name__ + ': \n' + self.__str__()


class ThrottleError(SdkError):
    """Service is busy"""

    def __init__(self):
        super(ThrottleError, self).__init__('service is busy, retry later')


class NetworkError(SdkError):
    """Network-level errors - connection error, timeout error, etc."""

    def __init__(self, extra=None):
        super(NetworkError, self).__init__('network error', None, extra)


class RequestError(SdkError):
    """Request-related errors - bad request, invalid payload, malformed transaction, etc."""

    def __init__(self, error_code=None, extra=None):
        super(RequestError, self).__init__('bad request', error_code, extra)


class ServerError(SdkError):
    """Server-related errors - rate limit exceeded, server over capacity."""

    def __init__(self, error_code=None, extra=None):
        super(ServerError, self).__init__('server error', error_code, extra)


class ResourceNotFoundError(SdkError):
    """Resource not found on the server."""

    def __init__(self, error_code=None, extra=None):
        super(ResourceNotFoundError, self).__init__('resource not found', error_code, extra)


class AccountError(SdkError):
    """Base class for account-related errors."""

    def __init__(self, address=None, message=None, error_code=None, extra=None):
        if address:
            extra = dict(extra or ())
            extra.update({'account': address})
        super(AccountError, self).__init__(message, error_code, extra)


class AccountNotFoundError(AccountError):
    """Operation referenced a nonexistent account."""

    def __init__(self, address=None, error_code=None, extra=None):
        super(AccountNotFoundError, self).__init__(address, 'account not found', error_code, extra)


class AccountActivatedError(AccountError):
    """Trying to activate an activated account."""

    def __init__(self, address=None, error_code=None, extra=None):
        super(AccountActivatedError, self).__init__(address, 'account already activated', error_code, extra)


class AccountExistsError(AccountError):
    """Trying to create an existing account."""

    def __init__(self, address=None, error_code=None, extra=None):
        super(AccountExistsError, self).__init__(address, 'account already exists', error_code, extra)


class AccountNotActivatedError(AccountError):
    """Operation referenced an account that exists but not yet activated."""

    def __init__(self, address=None, error_code=None, extra=None):
        super(AccountNotActivatedError, self).__init__(address, 'account not activated', error_code, extra)


class LowBalanceError(SdkError):
    """Account balance is too low to complete the operation. Refers both to native and asset balance."""

    def __init__(self, error_code=None, extra=None):
        super(LowBalanceError, self).__init__('low balance', error_code, extra)


class InternalError(SdkError):
    """Internal unhandled error. To find out more, check the error code and extra data."""

    def __init__(self, error_code=None, extra=None):
        super(InternalError, self).__init__('internal error', error_code, extra)


class FriendbotError(SdkError):
    """Friendbot related error"""

    def __init__(self, error_code=None, extra=None):
        super(FriendbotError, self).__init__('friendbot error', error_code, extra)


class CantSimplifyError(SdkError):
    """Transaction is too complex to simplify"""

    def __init__(self, error_code=None, extra=None):
        super(CantSimplifyError, self).__init__('Tx simplification error', error_code, extra)


class WrongNetworkError(SdkError):
    """The account is not using the network specified in the tx"""

    def __init__(self, error_code=None, extra=None):
        super(WrongNetworkError, self).__init__('The account is not using the network specified in the transaction',
                                                error_code, extra)


def translate_error(err):
    """A high-level error translator."""
    if isinstance(err, ClientError):
        return NetworkError({'internal_error': str(err)})
    if isinstance(err, HorizonError):
        return translate_horizon_error(err)
    return InternalError(None, {'internal_error': str(err)})


def translate_horizon_error(horizon_error):
    """Horizon error translator."""
    # query errors
    if horizon_error.type == HorizonErrorType.BAD_REQUEST:
        return RequestError(horizon_error.type, {'invalid_field': horizon_error.extras.get('invalid_field')})
    if horizon_error.type == HorizonErrorType.NOT_FOUND:
        return ResourceNotFoundError(horizon_error.type)
    if horizon_error.type in [HorizonErrorType.FORBIDDEN,
                              HorizonErrorType.NOT_ACCEPTABLE,
                              HorizonErrorType.UNSUPPORTED_MEDIA_TYPE,
                              HorizonErrorType.NOT_IMPLEMENTED,
                              HorizonErrorType.BEFORE_HISTORY,
                              HorizonErrorType.STALE_HISTORY]:
        return RequestError(horizon_error.type)

    # transaction (submit) errors
    if horizon_error.type == HorizonErrorType.TRANSACTION_MALFORMED:
        return RequestError(horizon_error.type)
    if horizon_error.type == HorizonErrorType.TRANSACTION_FAILED:
        return translate_transaction_error(horizon_error)

    # server errors
    if horizon_error.type == HorizonErrorType.RATE_LIMIT_EXCEEDED \
            or horizon_error.type == HorizonErrorType.SERVER_OVER_CAPACITY \
            or horizon_error.type == HorizonErrorType.TIMEOUT:
        return ServerError(horizon_error.type)
    if horizon_error.type == HorizonErrorType.INTERNAL_SERVER_ERROR:
        return InternalError(horizon_error.type)

    # unknown
    return InternalError(horizon_error.type, {'internal_error': 'unknown horizon error'})


def translate_transaction_error(tx_error):
    """Transaction error translator."""
    tx_result_code = tx_error.extras['result_codes']['transaction']
    if tx_result_code in [TransactionResultCode.TOO_EARLY,
                          TransactionResultCode.TOO_LATE,
                          TransactionResultCode.MISSING_OPERATION,
                          TransactionResultCode.BAD_AUTH,
                          TransactionResultCode.BAD_AUTH_EXTRA,
                          TransactionResultCode.BAD_SEQUENCE,
                          TransactionResultCode.INSUFFICIENT_FEE]:
        return RequestError(tx_result_code)
    if tx_result_code == TransactionResultCode.NO_ACCOUNT:
        return AccountNotFoundError(error_code=tx_result_code)
    if tx_result_code == TransactionResultCode.INSUFFICIENT_BALANCE:
        return LowBalanceError(tx_result_code)
    if tx_result_code == TransactionResultCode.FAILED:
        return translate_operation_error(tx_error.extras['result_codes']['operations'])
    return InternalError(tx_result_code, {'internal_error': 'unknown transaction error'})


def translate_operation_error(op_result_codes):
    """Operation error translator."""
    # Find first failed op
    for code in op_result_codes:
        if code != OperationResultCode.SUCCESS:
            op_result_code = code
            break
    if op_result_code in [OperationResultCode.BAD_AUTH,
                          CreateAccountResultCode.MALFORMED,
                          PaymentResultCode.NO_ISSUER,
                          PaymentResultCode.LINE_FULL,
                          ChangeTrustResultCode.INVALID_LIMIT]:
        return RequestError(op_result_code)
    if op_result_code in [OperationResultCode.NO_ACCOUNT, PaymentResultCode.NO_DESTINATION]:
        return AccountNotFoundError(error_code=op_result_code)
    if op_result_code == CreateAccountResultCode.ACCOUNT_EXISTS:
        return AccountExistsError(error_code=op_result_code)
    if op_result_code in [CreateAccountResultCode.LOW_RESERVE, PaymentResultCode.UNDERFUNDED]:
        return LowBalanceError(op_result_code)
    if op_result_code in [PaymentResultCode.SRC_NO_TRUST,
                          PaymentResultCode.NO_TRUST,
                          PaymentResultCode.SRC_NOT_AUTHORIZED,
                          PaymentResultCode.NOT_AUTHORIZED]:
        return AccountNotActivatedError(error_code=op_result_code)
    return InternalError(op_result_code, {'internal_error': 'unknown operation error'})



"""Contains the Environment class to configure horizon"""

from hashlib import sha256

from kin_base.network import NETWORKS

from typing import Optional


class Environment:
    def __init__(self, name: str, horizon_endpoint_uri: str, network_passphrase: str,
                 friendbot_url: Optional[str] = None):
        """
        Environments holds the parameters that will be used to connect to horizon

        :param name: Name of the environment.
        :param horizon_endpoint_uri: a Horizon endpoint.
        :param network_passphrase: The passphrase/network_id of the environment.
        :param friendbot_url: a url to a friendbot service
        :return: An instance of the Environment class.
        """
        # Add the network to the kin_base network list.
        NETWORKS[name.upper()] = network_passphrase
        self.name = name.upper()
        self.horizon_uri = horizon_endpoint_uri
        self.friendbot_url = friendbot_url
        self.passphrase = network_passphrase

        # Calculate the hash of the passphrase, can be used to calculate tx hash.
        self.passphrase_hash = sha256(network_passphrase.encode()).digest()

    def __str__(self):
        string_representation = '<Kin Environment>: Name: {}, ' \
                                'Horizon: {}, ' \
                                'Passphrase: {}'.format(self.name, self.horizon_uri, self.passphrase)
        return string_representation



"""Contains useful methods to be used across the sdk"""

from kin_base.utils import decode_check


def is_valid_address(address):
    """Determines if the provided string is a valid Stellar address.

    :param str address: address to check

    :return: True if this is a correct address
    :rtype: boolean
    """
    if len(address) != 56:
        return False

    try:
        decode_check('account', address)
        return True
    except:
        return False


def is_valid_secret_key(key):
    """Determines if the provided string is a valid Stellar key (seed).

    :param str key: key to check

    :return: True if this is a correct seed
    :rtype: boolean
    """
    if len(key) != 56:
        return False

    try:
        decode_check('seed', key)
        return True
    except:
        return False


def is_valid_transaction_hash(tx_hash):
    """Determines if the provided string is a valid Stellar transaction hash.

    :param str tx_hash: transaction hash to check

    :return: True if this is a correct transaction hash
    :rtype: boolean
    """
    if len(tx_hash) != 64:
        return False

    try:
        int(tx_hash, 16)
        return True
    except:
        return False



from .config import TEST_ENVIRONMENT, PROD_ENVIRONMENT, SDK_USER_AGENT
from .client import KinClient
from .account import KinAccount
from . import errors as KinErrors
from .transactions import OperationTypes, decode_transaction
from .blockchain.keypair import Keypair
from .blockchain.environment import Environment

# Override kin_base user agent with the kin-sdk user agent
from kin_base import horizon
horizon.HEADERS['User-Agent'] = SDK_USER_AGENT



"""Contains errors types related to horizon"""

HORIZON_NS_PREFIX = 'https://stellar.org/horizon-errors/'
"""
Horizon error example:

{
    'status': 400,
    'title': 'Transaction Failed',
    'detail': 'The transaction failed when submitted to the stellar network. The `extras.result_codes` field on this '
              'response contains further details.  Descriptions of each code can be found at: '
              'https://www.stellar.org/developers/learn/concepts/list-of-operations.html',
    'instance': '903d29404b0e/DAurVuBoL4-004368',
    'extras': 	{
        'result_codes': {
            'operations': ['op_no_destination'],
            'transaction': 'tx_failed'
        }, 
        'envelope_xdr': u'AAAAAJgXswhWU+pdHmHIurQuHk4ziNlKFxEJltbMOpF6EqETAAAAZAAAed0AAAAQAAAAAAAAAAAAAAABAAAAAAAAA'
                        u'AEAAAAAxbIcFBPzPZbzjWdkSB5FCSIva+WdQ2Oi70GUmFvFmOcAAAABVEVTVAAAAAD284i665ald1Kiq064FGlL+'
                        u'Aeych/b9UQngBHR37ZeiwAAAAAF9eEAAAAAAAAAAAF6EqETAAAAQN5x3xaOaeDS5EF3tE0X9zXymhqkOg95Tyfgu'
                        u'//TCbv9XN49CHoH5K+BUH04o1ZAZdHbnBABxh44bu7zbFLgQQU=',
        'invalid_field': None,
        'result_xdr': u'AAAAAAAAAGT/////AAAAAQAAAAAAAAAB////+wAAAAA='
    },
    'type': 'https://stellar.org/horizon-errors/transaction_failed'
}

"""

# noinspection PyClassHasNoInit
class HorizonErrorType:
    BAD_REQUEST = 'bad_request'  # cannot understand the request due to invalid parameters
    BEFORE_HISTORY = 'before_history'  # outside the range of recorded history
    FORBIDDEN = 'forbidden'  # not authorized to see
    NOT_ACCEPTABLE = 'not_acceptable'  # cannot reply with the requested data format
    NOT_FOUND = 'not_found'  # resource not found
    NOT_IMPLEMENTED = 'not_implemented'  # request method is not supported
    RATE_LIMIT_EXCEEDED = 'rate_limit_exceeded'  # too many requests in a one hour time frame
    SERVER_OVER_CAPACITY = 'server_over_capacity'  # server is currently overloaded
    STALE_HISTORY = 'stale_history'  # historical request out of date than the configured threshold
    TIMEOUT = 'timeout'  # request timed out before completing
    TRANSACTION_MALFORMED = 'transaction_malformed'
    TRANSACTION_FAILED = 'transaction_failed'  # transaction well-formed but failed
    UNSUPPORTED_MEDIA_TYPE = 'unsupported_media_type'  # unsupported content type
    INTERNAL_SERVER_ERROR = 'server_error'


# references:
# - github.com/stellar/go/blob/master/xdr/xdr_generated.go
# - github.com/stellar/go/blob/master/services/horizon/internal/actions_transaction.go
# - github.com/stellar/go/blob/master/services/horizon/internal/render/problem/main.go
# - github.com/stellar/horizon/blob/master/src/github.com/stellar/horizon/codes/main.go

# noinspection PyClassHasNoInit
class TransactionResultCode:
    SUCCESS = 'tx_success'  # all operations succeeded
    FAILED = 'tx_failed'  # one of the operations failed (none were applied)
    TOO_EARLY = 'tx_too_early'  # ledger closeTime before minTime
    TOO_LATE = 'tx_too_late'  # ledger closeTime after maxTime
    MISSING_OPERATION = 'tx_missing_operation'  # no operation was specified
    BAD_SEQUENCE = 'tx_bad_seq'  # sequence number does not match source account
    BAD_AUTH = 'tx_bad_auth'  # too few valid signatures / wrong network
    INSUFFICIENT_BALANCE = 'tx_insufficient_balance'  # fee would bring account below reserve
    NO_ACCOUNT = 'tx_no_source_account'  # source account not found
    INSUFFICIENT_FEE = 'tx_insufficient_fee'  # fee is too small
    BAD_AUTH_EXTRA = 'tx_bad_auth_extra'  # unused signatures attached to transaction
    INTERNAL_ERROR = 'tx_internal_error'  # an unknown error occurred


# noinspection PyClassHasNoInit
class OperationResultCode:
    INNER = 'op_inner'
    BAD_AUTH = 'op_bad_auth'
    NO_ACCOUNT = 'op_no_source_account'
    NOT_SUPPORTED = 'op_not_supported'
    SUCCESS = 'op_success'


# noinspection PyClassHasNoInit
class CreateAccountResultCode:
    SUCCESS = 'op_success'  # account was created
    MALFORMED = 'op_malformed'  # invalid destination account
    UNDERFUNDED = 'op_underfunded'  # not enough funds in source account
    LOW_RESERVE = 'op_low_reserve'  # would create an account below the min reserve
    ACCOUNT_EXISTS = 'op_already_exists'  # account already exists


# noinspection PyClassHasNoInit
class PaymentResultCode:
    SUCCESS = 'op_success'  # payment successfully completed
    MALFORMED = 'op_malformed'  # bad input
    UNDERFUNDED = 'op_underfunded'  # not enough funds in source account
    SRC_NO_TRUST = 'op_src_no_trust'  # no trust line on source account
    SRC_NOT_AUTHORIZED = 'op_src_not_authorized'  # source not authorized to transfer
    NO_DESTINATION = 'op_no_destination'  # destination account does not exist
    NO_TRUST = 'op_no_trust'  # destination missing a trust line for asset
    NOT_AUTHORIZED = 'op_not_authorized'  # destination not authorized to hold asset
    LINE_FULL = 'op_line_full'  # destination would go above their limit
    NO_ISSUER = 'op_no_issuer'  # missing issuer on asset


# noinspection PyClassHasNoInit
class PathPaymentResultCode(PaymentResultCode):
    TOO_FEW_OFFERS = 'op_too_few_offers'  # not enough offers to satisfy path
    OFFER_CROSS_SELF = 'op_cross_self'  # would cross one of its own offers
    OVER_SOURCE_MAX = 'op_over_source_max'  # could not satisfy sendmax


# noinspection PyClassHasNoInit
class ManageOfferResultCode:
    SUCCESS = 'op_success'  # operation successful
    MALFORMED = 'op_malformed'  # generated offer would be invalid
    UNDERFUNDED = 'op_underfunded'  # doesn't hold what it's trying to sell
    LINE_FULL = 'op_line_full'  # can't receive more of what it's buying
    SELL_NO_TRUST = 'op_sell_no_trust'  # no trust line for what we're selling
    BUY_NO_TRUST = 'op_buy_no_trust'  # no trust line for what we're buying
    SELL_NOT_AUTHORIZED = 'sell_not_authorized'  # not authorized to sell
    BUY_NOT_AUTHORIZED = 'buy_not_authorized'  # not authorized to buy
    OFFER_CROSS_SELF = 'op_cross_self'  # would cross an offer from the same user
    SELL_NO_ISSUER = 'op_sell_no_issuer'  # no issuer for what we're selling
    BUY_NO_ISSUER = 'op_buy_no_issuer'  # no issuer for what we're buying
    OFFER_NOT_FOUND = 'op_offer_not_found'  # offerID does not match an existing offer
    OFFER_LOW_RESERVE = 'op_low_reserve'  # not enough funds to create a new Offer


# noinspection PyClassHasNoInit
class SetOptionsResultCode:
    SUCCESS = 'op_success'  # operation successful
    LOW_RESERVE = 'op_low_reserve'  # not enough funds to add a signer
    TOO_MANY_SIGNERS = 'op_too_many_signers'  # max number of signers already reached
    BAD_FLAGS = 'op_bad_flags'  # invalid combination of clear/set flags
    INVALID_INFLATION = 'op_invalid_inflation'  # inflation account does not exist
    CANT_CHANGE = 'op_cant_change'  # can no longer change this option
    UNKNOWN_FLAG = 'op_unknown_flag'  # can't set an unknown flag
    THRESHOLD_OUT_OF_RANGE = 'op_threshold_out_of_range'  # bad value for weight/threshold
    BAD_SIGNER = 'op_bad_signer'  # signer cannot be masterkey
    INVALID_HOME_DOMAIN = 'op_invalid_home_domain'  # malformed home domain


# noinspection PyClassHasNoInit
class ChangeTrustResultCode:
    SUCCESS = 'op_success'  # operation successful
    MALFORMED = 'op_malformed'  # bad input
    NO_ISSUER = 'op_no_issuer'  # could not find issuer
    LOW_RESERVE = 'op_low_reserve'  # not enough funds to create a new trust line
    INVALID_LIMIT = 'op_invalid_limit'  # cannot drop limit below balance, cannot create with a limit of 0


# noinspection PyClassHasNoInit
class AllowTrustResultCode:
    SUCCESS = 'op_success'  # operation successful
    MALFORMED = 'op_malformed'  # asset is not ASSET_TYPE_ALPHANUM
    NO_TRUST_LINE = 'op_no_trustline'  # trustor does not have a trustline
    NOT_REQUIRED = 'op_not_required'  # source account does not require trust
    CANT_REVOKE = 'op_cant_revoke'  # source account can't revoke trust


# noinspection PyClassHasNoInit
class AccountMergeResultCode:
    SUCCESS = 'op_success'  # operation successful
    MALFORMED = 'op_malformed'  # can't merge onto itself
    NO_ACCOUNT = 'op_no_account'  # destination does not exist
    IMMUTABLE_SET = 'op_immutable_set'  # source account has AUTH_IMMUTABLE set
    HAS_SUB_ENTRIES = 'op_has_sub_entries'  # account has trust lines/offers


# noinspection PyClassHasNoInit
class InflationResultCode:
    SUCCESS = 'op_success'
    NOT_TIME = 'op_not_time'
